"""Command line tool for diagnosing configuration problems in the cluster."""

import logging
from argparse import ArgumentParser, _SubParsersAction as SubParsersAction
from typing import cast
import os
import pathlib
import yaml


_LOGGER = logging.getLogger(__name__)

FAIL = "[DIAGNOSTICS FAIL]"
OK = "[DIAGNOSTICS OK]"


class DiagnosticsAction:
    """Flux-local diagnostics action."""

    @classmethod
    def register(
        cls, subparsers: SubParsersAction  # type: ignore[type-arg]
    ) -> ArgumentParser:
        """Register the subparser commands."""
        args = cast(
            ArgumentParser,
            subparsers.add_parser(
                "diagnostics",
                help="Print information about local flux that can diagnose issues",
                description="Print information about the local cluster to aid with diagnosing problems.",
            ),
        )
        args.add_argument(
            "--path",
            help="Optional path with flux Kustomization resources (multi-cluster ok)",
            type=pathlib.Path,
            default=None,
            nargs="?",
        )
        args.set_defaults(cls=cls)
        return args

    async def run(  # type: ignore[no-untyped-def]
        self,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        path = kwargs.get("path") or "."

        # Parse directories to ignore from `.krmignore`
        krmignore = pathlib.Path(path) / ".krmignore"
        ignoredirs = []
        if krmignore.exists():
            with krmignore.open() as fd:
                ignoredirs = [
                    str(pathlib.Path(line.strip())) for line in fd.readlines()
                ]

        errors = []
        for root, dirs, files in os.walk(str(path)):
            # Ignore any directories that we should not walk
            rootpath = str(pathlib.Path(root))
            if any([rootpath.startswith(i) for i in ignoredirs]):
                continue

            for file in files:
                if not (file.endswith(".yaml") or file.endswith(".yml")):
                    continue

                full_path = pathlib.Path(root) / file
                try:
                    doc = list(yaml.safe_load_all(full_path.read_text()))
                except yaml.YAMLError as err:
                    errors.append(f"`{full_path}` failed to parse as yaml: {err}")
                    continue

                for subdoc in doc:
                    if isinstance(subdoc, dict):
                        continue
                    if isinstance(subdoc, list):
                        errors.append(
                            f"`{full_path}` expected dictionary but was list: {subdoc}"
                        )
                        break
                    errors.append(
                        f"`{full_path}` was not a dictionary: {type(subdoc)}: {subdoc}"
                    )

        if errors:
            for error in errors:
                print(f"{FAIL}: {error}")
        else:
            print(OK)



"""Test helpers for flux-local tools."""

from flux_local.command import Command, run

FLUX_LOCAL_BIN = "flux-local"


async def run_command(args: list[str], env: dict[str, str] | None = None) -> str:
    return await run(Command([FLUX_LOCAL_BIN] + args, env=env))



"""Tests for manifest library."""

from pathlib import Path

import pytest
import yaml

from flux_local.manifest import (
    Cluster,
    HelmRelease,
    HelmRepository,
    Manifest,
    read_manifest,
    write_manifest,
    NamedResource,
)

TESTDATA_DIR = Path("tests/testdata/cluster/infrastructure")
TEST_PODINFO_HELMRELEASE = Path("tests/testdata/cluster8/apps/podinfo.yaml")


def test_parse_helm_release() -> None:
    """Test parsing a helm release doc."""

    release = HelmRelease.parse_doc(
        yaml.load(
            (TESTDATA_DIR / "controllers/metallb-release.yaml").read_text(),
            Loader=yaml.CLoader,
        )
    )
    assert release.name == "metallb"
    assert release.namespace == "metallb"
    assert release.chart.name == "metallb"
    assert release.chart.version == "4.1.14"
    assert release.chart.repo_name == "bitnami"
    assert release.chart.repo_namespace == "flux-system"
    assert release.values


def test_compact_helm_release() -> None:
    """Test parsing a helm release doc."""

    release = HelmRelease.parse_doc(
        yaml.load(
            (TESTDATA_DIR / "controllers/metallb-release.yaml").read_text(),
            Loader=yaml.CLoader,
        )
    )
    print(release)
    assert release.compact_dict() == {
        "name": "metallb",
        "namespace": "metallb",
        "chart": {
            "name": "metallb",
            "repo_kind": "HelmRepository",
            "repo_name": "bitnami",
            "repo_namespace": "flux-system",
        },
    }


def test_parse_helm_repository() -> None:
    """Test parsing a helm repository doc."""

    docs = list(
        yaml.load_all(
            (TESTDATA_DIR / "configs/helm-repositories.yaml").read_text(),
            Loader=yaml.CLoader,
        )
    )
    assert len(docs) == 3
    repo = HelmRepository.parse_doc(docs[0])
    assert repo.name == "bitnami"
    assert repo.namespace == "flux-system"
    assert repo.url == "https://charts.bitnami.com/bitnami"
    assert repo.repo_type == "default"
    repo = HelmRepository.parse_doc(docs[1])
    assert repo.name == "podinfo"
    assert repo.namespace == "flux-system"
    assert repo.url == "oci://ghcr.io/stefanprodan/charts"
    assert repo.repo_type == "oci"


async def test_write_manifest_file() -> None:
    """Test reading an invalid manifest file."""
    await write_manifest(Path("/dev/null"), Manifest(clusters=[]))


async def test_read_write_empty_manifest(tmp_path: Path) -> None:
    """Test serializing and reading back a manifest."""
    manifest = Manifest(clusters=[])
    await write_manifest(tmp_path / "file.yaml", manifest)
    new_manifest = await read_manifest(tmp_path / "file.yaml")
    assert not new_manifest.clusters


async def test_read_manifest_invalid_file() -> None:
    """Test reading an invalid manifest file."""
    with pytest.raises(ValueError, match="validation error for Manifest"):
        await read_manifest(Path("/dev/null"))


async def test_serializing_manifest(tmp_path: Path) -> None:
    """Test serializing a manifest to a dictionary."""
    manifest = Manifest(
        clusters=[
            Cluster(
                path="./example",
                kustomizations=[],
            )
        ]
    )
    await write_manifest(tmp_path / "file.yaml", manifest)
    new_manifest = await read_manifest(tmp_path / "file.yaml")
    assert new_manifest.compact_dict() == {
        "clusters": [
            {
                "path": "./example",
                "kustomizations": [],
            },
        ]
    }


def test_parse_helmrelease_chartref() -> None:
    """Test parsing a helm release doc."""

    HELM_CHARTREF_FILE = Path("tests/testdata/cluster9/apps/podinfo/podinfo.yaml")
    docs = list(
        yaml.load_all(
            HELM_CHARTREF_FILE.read_text(),
            Loader=yaml.CLoader,
        )
    )
    assert len(docs) == 1
    assert docs[0].get("kind") == "HelmRelease"

    release = HelmRelease.parse_doc(docs[0])
    assert release.name == "podinfo"
    assert release.namespace == "default"
    assert release.chart.name == "podinfo"
    assert release.chart.version is None
    assert release.chart.repo_name == "podinfo"
    assert release.chart.repo_namespace == "default"
    assert release.values


def test_helmrelease_dependencies() -> None:
    """Test parsing a helm release doc."""

    docs = list(
        yaml.load_all(
            TEST_PODINFO_HELMRELEASE.read_text(),
            Loader=yaml.CLoader,
        )
    )
    assert len(docs) == 2
    release = HelmRelease.parse_doc(docs[1])
    assert release.name == "podinfo"
    assert release.namespace == "podinfo"
    # Assert on all the input dependencies that might cause HelmRelease
    # output to change. This means the HelmRepository, and any input
    # values used for substitute references.
    assert release.resource_dependencies == [
        NamedResource(kind="HelmRelease", name="podinfo", namespace="podinfo"),
        NamedResource(kind="HelmRepository", name="podinfo", namespace="flux-system"),
        NamedResource(kind="ConfigMap", name="podinfo-values", namespace="podinfo"),
        NamedResource(kind="Secret", name="podinfo-tls-values", namespace="podinfo"),
    ]



"""Module for working with kubernetes values."""

import base64
from collections.abc import Iterable, Generator, Callable
import logging
from typing import TypeVar, Any

import yaml

from .manifest import (
    HelmRelease,
    SECRET_KIND,
    Kustomization,
    CONFIG_MAP_KIND,
    ConfigMap,
    Secret,
    VALUE_PLACEHOLDER_TEMPLATE,
    ValuesReference,
)
from .exceptions import HelmException, InputException, InvalidValuesReference


_LOGGER = logging.getLogger(__name__)


_T = TypeVar("_T", bound=ConfigMap | Secret)


class ClusterConfig:
    """Interface for accessing all cluster configuration."""

    def __init__(
        self,
        secrets: Callable[[], Iterable[Secret]],
        config_maps: Callable[[], Iterable[ConfigMap]],
    ) -> None:
        """Initialize ClusterConfig."""
        self._secrets = secrets
        self._config_maps = config_maps

    @property
    def secrets(self) -> Iterable[Secret]:
        """Available Secret objects in the cluster."""
        return iter(self._secrets())

    @property
    def config_maps(self) -> Iterable[ConfigMap]:
        """Available ConfigMap objects in the cluster."""
        return iter(self._config_maps())


def cluster_config(
    secrets: list[Secret], config_maps: list[ConfigMap]
) -> ClusterConfig:
    """Create a ClusterConfig from a list of secrets and configmaps."""
    return ClusterConfig(
        lambda: secrets,
        lambda: config_maps,
    )


def merge_cluster_config(
    config: ClusterConfig, secrets: list[Secret], config_maps: list[ConfigMap]
) -> ClusterConfig:
    """Create a ClusterConfig from a list of secrets and configmaps."""
    return ClusterConfig(
        lambda: list(config.secrets) + secrets,
        lambda: list(config.config_maps) + config_maps,
    )


def ks_cluster_config(kustomizations: list[Kustomization]) -> ClusterConfig:
    """Create a ClusterConfig from a list of Kustomizations."""

    def secrets_iter() -> Generator[Secret, None, None]:
        for ks in kustomizations:
            for secret in ks.secrets:
                yield secret

    def config_maps_iter() -> Generator[ConfigMap, None, None]:
        for ks in kustomizations:
            for config_map in ks.config_maps:
                yield config_map

    return ClusterConfig(
        secrets_iter,
        config_maps_iter,
    )


def _find_object(name: str, namespace: str, objects: Iterable[_T]) -> _T | None:
    """Find the object in the list of objects."""
    for obj in objects:
        if obj.name == name and obj.namespace == namespace:
            return obj
    return None


def _decode_config_or_secret_value(
    name: str, string_data: dict[str, str] | None, binary_data: dict[str, str] | None
) -> dict[str, str] | None:
    """Return the config or secret data."""
    if binary_data:
        try:
            return {
                k: base64.b64decode(v).decode("utf-8") for k, v in binary_data.items()
            }
        except ValueError:
            raise HelmException(f"Unable to decode binary data for configmap {name}")
    return string_data


def _get_secret_data(
    name: str, namespace: str, secrets: Iterable[Secret]
) -> dict[str, str] | None:
    """Find the secret value in the kustomization."""
    found: Secret | None = _find_object(name, namespace, secrets)
    if not found:
        return None
    return _decode_config_or_secret_value(
        f"{namespace}/{name}", found.string_data, found.data
    )


def _get_configmap_data(
    name: str, namespace: str, config_maps: Iterable[ConfigMap]
) -> dict[str, str] | None:
    """Find the configmap value in the kustomization."""
    found: ConfigMap | None = _find_object(name, namespace, config_maps)
    if not found:
        return None
    return _decode_config_or_secret_value(
        f"{namespace}/{name}", found.data, found.binary_data
    )


def _lookup_value_reference(
    ref: ValuesReference,
    namespace: str,
    cluster_config: ClusterConfig,
) -> str | None:
    _LOGGER.debug("Expanding value reference %s", ref)
    found_data: dict[str, str] | None = None
    if ref.kind == SECRET_KIND:
        found_data = _get_secret_data(ref.name, namespace, cluster_config.secrets)
    elif ref.kind == CONFIG_MAP_KIND:
        found_data = _get_configmap_data(
            ref.name, namespace, cluster_config.config_maps
        )
    else:
        raise InvalidValuesReference(f"Unsupported valueFrom kind {ref.kind}")

    found_value: str | None = None
    if found_data is None:
        if not ref.optional:
            _LOGGER.warning(
                "Unable to find %s %s/%s referenced",
                ref.kind,
                namespace,
                ref.name,
            )
        if ref.target_path:
            # When a target path is specified, the value is expected to be
            # a simple value type. Create a synthetic placeholder value, otherwise
            # there is nothing to replace.
            return VALUE_PLACEHOLDER_TEMPLATE.format(name=ref.name)
        return None

    elif (found_value := found_data.get(ref.values_key)) is None:
        raise InvalidValuesReference(
            f"Unable to find key {ref.values_key} in {namespace}/{ref.name}"
        )

    return found_value


def _deep_merge(base: dict[str, Any], override: dict[str, Any]) -> dict[str, Any]:
    """Recursively merge two dictionaries, similar to how Helm merges values. Lists are replaced entirely (Helm behavior)."""
    result = base.copy()
    for key, override_value in override.items():
        base_value = result.get(key)
        if (
            base_value is not None
            and isinstance(base_value, dict)
            and isinstance(override_value, dict)
        ):
            result[key] = _deep_merge(base_value, override_value)
        else:
            result[key] = override_value
    return result


def _update_helmrelease_values(
    ref: ValuesReference,
    found_value: str,
    values: dict[str, Any],
) -> dict[str, Any]:
    """Expand value references in the HelmRelease."""

    if ref.target_path:
        parts = ref.target_path.split(".")
        inner_values = values
        for part in parts[:-1]:
            if part not in inner_values:
                inner_values[part] = {}
            elif not isinstance(inner_values[part], dict):
                raise InputException(
                    f"Expected '{ref.name}' field '{ref.target_path}' values to be a dict, found {type(inner_values[part])}"
                )
            inner_values = inner_values[part]

        inner_values[parts[-1]] = found_value
    else:
        obj = yaml.load(found_value, Loader=yaml.SafeLoader)
        # Handle empty YAML file case
        if obj is None:
            obj = {}
        if not isinstance(obj, dict):
            raise InputException(
                f"Expected '{ref.name}' field '{ref.target_path}' values to be valid yaml, found {type(values)}"
            )
        values = _deep_merge(values, obj)

    return values


def expand_value_references(
    helm_release: HelmRelease, kustomization: Kustomization
) -> HelmRelease:
    if not helm_release.values_from:
        return helm_release

    values: dict[str, Any] = {}
    cluster_config = ks_cluster_config([kustomization])
    for ref in helm_release.values_from:
        _LOGGER.debug("Expanding value reference %s", ref)

        try:
            found_value = _lookup_value_reference(
                ref,
                helm_release.namespace,
                cluster_config,
            )
        except InvalidValuesReference as err:
            _LOGGER.warning(
                "Skipping ValuesReference for %s: %s", helm_release.namespaced_name, err
            )
            continue

        if found_value is None:
            continue

        try:
            values = _update_helmrelease_values(ref, found_value, values)
        except InputException as err:
            raise HelmException(
                f"Error building HelmRelease '{helm_release.namespaced_name}': {err}"
            )

    if helm_release.values:
        values = _deep_merge(values, helm_release.values)

    helm_release.values = values
    return helm_release


def expand_postbuild_substitute_reference(
    ks: Kustomization, cluster_config: ClusterConfig
) -> Kustomization:
    """Expand postbuild substituteFrom references to substitutions."""
    if not ks.postbuild_substitute_from:
        return ks

    values: dict[str, Any] = ks.postbuild_substitute or {}
    for ref in ks.postbuild_substitute_from:
        _LOGGER.debug("Expanding substitute reference %s", ref)
        if not ks.namespace:
            raise InvalidValuesReference(
                "Kustomization with substitueFrom has no namespace"
            )
        found_data: dict[str, str] | None = None
        if ref.kind == SECRET_KIND:
            found_data = _get_secret_data(
                ref.name, ks.namespace, cluster_config.secrets
            )
        elif ref.kind == CONFIG_MAP_KIND:
            found_data = _get_configmap_data(
                ref.name, ks.namespace, cluster_config.config_maps
            )
        else:
            _LOGGER.warning(
                "Skipping SubstituteReference for %s: %s", ks.namespaced_name, ref.name
            )
            continue

        if found_data is None:
            if (
                not ref.optional and not ref.kind == SECRET_KIND
            ):  # Secrets are commonly filtered
                _LOGGER.warning(
                    "Unable to find SubstituteReference for %s: %s",
                    ks.namespaced_name,
                    ref.name,
                )
            continue

        values.update(found_data)
    _LOGGER.debug("update_postbuild_substitutions=%s", values)
    ks.update_postbuild_substitutions(values)
    return ks



"""Tests for the flux-local `diff ks` command."""

import pytest

from syrupy.assertion import SnapshotAssertion

from . import run_command


@pytest.mark.parametrize(
    ("args", "env"),
    [
        (["apps", "--path", "tests/testdata/cluster"], {}),
        (["apps", "--path", "tests/testdata/cluster"], {"DIFF": "diff"}),
        (["apps", "--path", "tests/testdata/cluster", "-o", "yaml"], {}),
        (
            [
                "apps",
                "--path",
                "tests/testdata/cluster",
                "-o",
                "yaml",
                "--limit-bytes",
                "20000",
            ],
            {},
        ),
        (
            [
                "apps",
                "--path",
                "tests/testdata/cluster",
                "-o",
                "yaml",
                "--sources",
                "''",
            ],
            {},
        ),
    ],
    ids=["apps", "ks-external", "yaml", "yaml-limit", "yaml-empty-sources"],
)
async def test_diff_ks(
    args: list[str], env: dict[str, str], snapshot: SnapshotAssertion
) -> None:
    """Test test diff ks commands."""
    result = await run_command(["diff", "ks"] + args, env=env)
    assert result == snapshot



"""Exceptions related to flux-local."""

__all__ = [
    "FluxException",
    "InputException",
    "CommandException",
]


class FluxException(Exception):
    """Generic base exception used for this library."""


class InputException(FluxException):
    """Raised when the input files or values are not formatted as expected."""


class CommandException(FluxException):
    """Raised when there is a failure running a subcommand."""


class KustomizeException(CommandException):
    """Raised when there is a failure running a kustomize command."""


class KustomizePathException(CommandException):
    """Raised a Kustomization points to a path that does not exist."""


class HelmException(CommandException):
    """Raised when there is a failure running a helm command."""


class InvalidValuesReference(FluxException):
    """Exception raised for an unsupported ValuesReference."""


class InvalidSubstituteReference(FluxException):
    """Exception raised for an unsupported SubstituteReference."""



"""Tests for the flux-local diff command.

This uses a separate test harness to create a worktree, modify files, then
run the diff. This is difficult to setup with static testdata since the
Kustomization paths would need to change which makes the resources
look like they belong to a separate cluster.
"""

import pytest
import os

from syrupy.assertion import SnapshotAssertion

from flux_local import git_repo

from . import run_command

CLUSTER_DIR = "tests/testdata/cluster"


# These tests fool codecov so run separately
@pytest.mark.skipif(
    os.environ.get("SKIP_DIFF_TESTS", False),
    reason="SKIP_DIFF_TESTS set in environment",
)
async def test_diff_ks(snapshot: SnapshotAssertion) -> None:
    """Test a diff in resources within a Kustomization."""

    repo = git_repo.git_repo()
    with git_repo.create_worktree(repo) as tree1:
        path_orig = tree1 / CLUSTER_DIR

        with git_repo.create_worktree(repo) as tree2:
            path = tree2 / CLUSTER_DIR

            # Empty out a config map in the cluster
            configmap = path / "apps/prod/configmap.yaml"
            configmap.write_text("")

            result = await run_command(
                [
                    "diff",
                    "ks",
                    "--path",
                    str(path),
                    "--path-orig",
                    str(path_orig),
                ]
            )
    assert result == snapshot



"""Tests for kustomize library."""

from pathlib import Path

import pytest
from syrupy.assertion import SnapshotAssertion
import yaml

from flux_local import kustomize, exceptions, manifest, command

TESTDATA_DIR = Path("tests/testdata")

KUSTOMIZATION = """---
apiVersion: kustomize.config.k8s.io/v1beta1
kind: Kustomization
resources:
- example.yaml
"""


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_grep(path: Path, snapshot: SnapshotAssertion) -> None:
    """Test a kustomize grep command."""
    result = await kustomize.grep("kind=ConfigMap", path).run()
    assert "Secret" not in result
    assert "ConfigMap" in result
    assert result == snapshot


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_objects(path: Path, snapshot: SnapshotAssertion) -> None:
    """Test loading yaml documents."""
    cmd = kustomize.grep("kind=ConfigMap", path)
    result = await cmd.objects()
    assert len(result) == 1
    assert result[0].get("kind") == "ConfigMap"
    assert result[0].get("apiVersion") == "v1"
    assert result == snapshot


INVALID_YAML = """
---
foo: !bar
"""


async def test_objects_failure() -> None:
    """Test loading yaml documents."""

    class FakeTask(command.Task):
        async def run(self, stdin: bytes | None = None) -> bytes:
            """Execute the task and return the result."""
            return INVALID_YAML.encode()

    cmd = kustomize.Kustomize([FakeTask()])
    with pytest.raises(
        exceptions.KustomizeException,
        match=r"Unable to parse.*could not determine a constructor",
    ):
        await cmd.objects()


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_stash(path: Path) -> None:
    """Test loading yaml documents."""
    cmd = await kustomize.grep("kind=Ignored", path, invert=True).stash()
    result = await cmd.grep("kind=ConfigMap").objects()
    assert len(result) == 1
    assert result[0].get("kind") == "ConfigMap"
    assert result[0].get("apiVersion") == "v1"
    result = await cmd.grep("kind=Secret").objects()
    assert len(result) == 1
    assert result[0].get("kind") == "Secret"
    assert result[0].get("apiVersion") == "v1"
    result = await cmd.grep("kind=Unknown").objects()
    assert len(result) == 0


async def test_target_namespace() -> None:
    """Test a kustomization with a target namespace."""
    ks = kustomize.grep("kind=ConfigMap", TESTDATA_DIR / "repo")

    result = await ks.objects()
    assert len(result) == 1
    config_map = result[0]
    assert "metadata" in config_map
    assert config_map["metadata"] == {
        "name": "cluster-settings",
        "namespace": "flux-system",
        "annotations": {
            "config.kubernetes.io/index": "0",
            "config.kubernetes.io/path": "cluster-settings.yaml",
            "internal.config.kubernetes.io/index": "0",
            "internal.config.kubernetes.io/path": "cluster-settings.yaml",
        },
    }

    result = await ks.objects(target_namespace="configs")
    assert len(result) == 1
    config_map = result[0]
    assert "metadata" in config_map
    assert config_map["metadata"] == {
        "name": "cluster-settings",
        # Verify updated namespace
        "namespace": "configs",
        "annotations": {
            "config.kubernetes.io/index": "0",
            "config.kubernetes.io/path": "cluster-settings.yaml",
            "internal.config.kubernetes.io/index": "0",
            "internal.config.kubernetes.io/path": "cluster-settings.yaml",
        },
    }


async def test_flux_build_path_is_not_dir() -> None:
    """Test case where the flux build path does not exist."""
    cmd = kustomize.flux_build(
        manifest.Kustomization(name="example", path="./", namespace="flux-system"),
        Path(TESTDATA_DIR) / "does-not-exist",
    )
    with pytest.raises(exceptions.FluxException, match="not a directory"):
        await cmd.objects()


async def test_flux_build() -> None:
    """Test flux build cli."""
    docs = list(
        yaml.safe_load_all(
            Path(
                f"{TESTDATA_DIR}/cluster/clusters/prod/flux-system/gotk-sync.yaml"
            ).read_text()
        )
    )
    assert len(docs) == 2
    ks = manifest.Kustomization.parse_doc(docs[1])
    cmd = kustomize.flux_build(ks, Path(ks.path))
    result = await cmd.run()
    assert "GitRepository" in result
